"""MicroMPN: a Python command line program for automating microbiology most probable 
   number (MPN) estimates in laboratory microplates

   United States Department of Agriculture

"""
import os
import sys
import argparse
import logging
from typing import Tuple, List, Dict

import numpy as np
import pandas as pd
import wellmap
import micrompn.mpn
from micrompn import _version


def myparser() -> argparse.ArgumentParser:
    """Parse the input arguments for MicroMPN 

    :return: _description_
    :rtype: argparse.ArgumentParser
    """

    parser = argparse.ArgumentParser(
    description='MicroMPN: Software to estimate Most Probable Number (MPN) bacterial abundance from microplates')
    parser.add_argument('--wellmap', type=str, required=True, help='A TOML file  with plate layout specified in wellmap format ')
    parser.add_argument('--data', type=str, required=True, help='A csv file or a directory containing csv files with the plate name, optical value, and well or row and column data')
    parser.add_argument('--cutoff', type=float, required=True, help='The value from the plate reader above which a well is classified as positive')
    parser.add_argument('--outfile', type=str, required=True, help='The file path and name for the results')
    parser.add_argument('--plate_name', type=str, default='plate', help='The name of the column containing the plate identifier in the data file')
    parser.add_argument('--value_name', type=str, default='rfu', help='The name of the column containing the optical signal column in the data file')
    parser.add_argument('--well_name', type=str, required=False, help='The name of the column containing the well identifier in the data file')
    parser.add_argument('--col_name', type=str, required=False, help='The name of the column containing the plate column identifier in the data file')
    parser.add_argument('--row_name', type=str, required=False, help='The name of the column containing the plate row identifier in the data file')   
    parser.add_argument('--zero_padded', action='store_true', help='If present, the well value in the data file is treated as zero-padded, e.g. A01')
    parser.add_argument('--trim_positives', action='store_true', help='If present, the list of positive wells will be trimmed to the most dilute  \
                                                                    all positive dilution and the least dilute all negative dilution. This helps if early dilutions are turbid. ')
    parser.add_argument('--version', '-v',  action='version', version="%(prog)s (" + _version.__version__ + ")")
    parser.add_argument('--logfile', '-l', type=str, default= "micrompn.log")
    return parser
    
def _logger_setup(logfile: str) -> logging.Logger:
    """Set up logger for micrompn

    :param logfile: The path for the logger
    :type logfile: str
    :raises e: A generic error if logger could not be set up
    :return: Returns a configured logger
    :rtype: logging.logger
    """
    try:
        # create logger
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        # create console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # create file handler

        fh = logging.FileHandler(logfile)
        fh.setLevel(logging.DEBUG)
        # set a format which is simpler for console use
        formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        # tell the handler to use this format
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(fh)
        logger.addHandler(ch)
        return logger
    except Exception as e:
        print("An error occurred setting up logging")
        raise e

logger = _logger_setup("micrompn.log")

def split_dataframe(df: pd.DataFrame) -> Tuple[List[pd.DataFrame], np.ndarray]:
    """ The function breaks a dataframe of plate data and layout data and into a Tuple containing a list of 
        dataframes split by plates and samples and a pd.Series of unique samples

    :param df: A dataframe containing a sample column
    :type df: pd.DataFrame
    :return:  A list of dataframes split by sample and a Pandas Series with the samples names 
    :rtype: Tuple[List[pd.DataFrame], pd.Series]
    """
    df['plate_sample'] = df["plate"] + '_' + df['sample'].astype(str)
    unique_samples = df['plate_sample'].unique() # Get unique values in the 'plate_sample' column
    list_of_dfs = []
    for sample in unique_samples:
        sample_df = df[df['plate_sample'] == sample] # Filter the dataframe by the unique sample
        list_of_dfs.append(sample_df) # Add the filtered dataframe to the list
    return list_of_dfs, unique_samples

def count_above_cutoff(df: pd.DataFrame, cutoff: float) -> np.array:
    """ Counts how many wells are above a cutoff value at each dilution. 


    :param df: The input dataframe containing dilution values
    :type df: pd.DataFrame
    :param cutoff: The value above which a well is considered positive
    :type cutoff: float
    :return: An array of the counts of positive wells at each dilution
    :rtype: np.array
    """
    print(df)
    grouped = df.groupby('dilution')
    counts = []
    for name, group in grouped:
        above_cutoff = np.sum(group['value'] > cutoff)
        counts.append(above_cutoff)
    counts.reverse()

    return np.array(counts)

def trim_counts(positive:np.array, tubes:np.array, dilution: np.array) -> Tuple[np.array, np.array, np.array]:
    """ Trim the data to the first all negative and first all positive dilutions.
        This is useful when turbidity in early dilutions reduces fluorescence.

    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :return: A tuple of trimmed np arrays: positive, tubes, and dilution
    :rtype: Tuple[np.array, np.array, np.array]
    """
    arr = positive / tubes
    ones = np.where(np.isclose(arr, 1))[0]
    zeros = np.where(np.isclose(arr, 0))[0]
    if len(ones) == 0 or len(zeros) == 0:
        logger.warning("An all positive or all negative dilution could not be found. Returning the untrimmed vector for this sample")
        return (positive, tubes, dilution)
    min_idx = ones[-1]
    max_idx = zeros[0] + 1
    return (positive[min_idx:max_idx], tubes[min_idx:max_idx], dilution[min_idx:max_idx])
        


def import_plate(platefile: str,
                 plate_name: str,
                 well_name: str,
                 col_name: str,
                 row_name: str,
                 value_name: str,
                 zero_padded: bool) -> pd.DataFrame:
    """ import a csv of column data from a plate reader and convert it to a uniform 
        format for merging with the  plate layout file. This works for 96 and 384 well plates but not 1536 wells

    :param platefile: A string of the path to the file.
    :type platefile: str
    :param plate_name: A column label with a unique plate name.
    :type plate_name: str
    :param well_name: A column label for the well name (e.g. A1 or A01), optional if col and row are provided.
    :type well_name: str
    :param col_name: A column label for the column name (e.g. 12) rownames should br provided too. optional if wellname is provided.
    :type col_name: str
    :param row_name: A column label for the row name (e.g. D) colnames should br provided too. optional if wellname is provided.
    :type row_name: str
    :param value_name: A column label for the optical measurement used to determine Growth in the well
    :type value_name: str
    :param zero_padded: Does well have leading zero in the column number, e.g.  A01 instead of A1
    :type zero_padded: bool
    :return: A uniform dataframe for merging with the plate layout dataframe.
    :rtype: pd.DataFrame
    """
    assert(well_name or (col_name and row_name)),"you need to specify the columns to use for microplate well or microplate column and row."
    try:
        plate = pd.read_csv(platefile)
        dtype_dict = {'plate' : None, 'well' : None, 'col' : None, 'row' : None, 'value' : None }
        cleanplate = pd.DataFrame(columns=dtype_dict.keys()).astype(dtype_dict)
        cleanplate['plate'] = plate[plate_name]
        cleanplate['value'] = plate[value_name]
        if col_name and row_name:
            cleanplate['col'] = plate[col_name]
            cleanplate['row'] = plate[row_name]
            cleanplate['well'] =  plate[row_name] +  plate[col_name].str()
        elif well_name:
            if zero_padded:
                cleanplate['well'] = plate[well_name].astype(str).str[0] + plate[well_name].astype(str).str[1:].int().str()
            else:
                cleanplate['well'] = plate[well_name]    
            cleanplate['row'] = plate[well_name].str[0]
            cleanplate['col'] = plate[well_name].str.slice(start=1).str.zfill(2)
        return cleanplate
    except Exception as e:
        logger.exception(e)
        raise e
    
def process_plate(wellmap_file: str, plate_df: pd.DataFrame, args: argparse.Namespace ) -> pd.DataFrame:
    """Read plate data and return a dataframe of MPN estimates and statistics for each sample

    :param wellmap_file: Path to the Wellmap layout TOML file
    :type wellmap_file: str
    :param plate_df: A pandas DataFrame of plate reader data from one or more plates
    :type plate_df: pd.DataFrame
    :param args: The input parameter arguments
    :type args: argparse.Namespace
    :return: A data frame of plate, sample, MPN and MPN statistics
    :rtype: pd.DataFrame
    """
    try:
        dtype_dict = {'plate': str, 'sample': str, 'mpn': float, 'mpn_adj': float, 'upper': float, 'lower': float, 'rarity':float}
        temp_df= pd.DataFrame(columns=dtype_dict.keys()).astype(dtype_dict)
        layout = wellmap.load(toml_path=wellmap_file)
        data = layout.merge(right=plate_df, on = "well", how='left')
        df_list, unique_samples = split_dataframe(data)
        for df, sample_no in zip(df_list, unique_samples):
            positive = count_above_cutoff(df, args.cutoff)
            dilution_list = (df['dilution'].unique())
            tubes = np.repeat(a = len(df['replicate'].unique()), repeats=len(dilution_list))
            if args.trim_positives:
                positive, tubes, dilution_list = trim_counts(positive, tubes, dilution_list)
            mobj = micrompn.mpn(positive, tubes, dilution_list) 
            temp_df.loc[len(temp_df)] = [list(df['plate'])[0], sample_no.split('_')[1], mobj['MPN'], mobj['MPN_adj'], mobj['UB'], mobj['LB'], mobj['RI']]
        return temp_df
    except Exception as e:
        logger.exception(e)
        raise e
    


def is_file_or_directory(path: str) -> bool:
    """check if a string is a path or a directory

    :param path: the file or director
    :type path: str
    :raises FileNotFoundError: Error is not a file or directory
    :return: true if a file, false if a directory
    :rtype: bool
    """
    try:
        if os.path.isfile(path):
            return True
        elif os.path.isdir(path):
            return False
    except FileNotFoundError(f"{path} not found") as e:
        logger.exception(e)
        raise e

def main(arglist: list = None) -> None:
    """An entry point to run the full micrompn workflow

    :param arglist: A list of arguments, used for testing, defaults to None
    :type arglist: list, optional
    """
    try:
        # Set up logging
        parser = myparser()
        args = parser.parse_args(arglist)
        logger = _logger_setup(args.logfile)
    except Exception as e:
        print("Could not parse input arguments and set up logging correctly")
        raise(e)
        sys.exit(1)
    # set up pandas datafreame 
    try:
        logger.info("Starting microMPN")
        logger.info("Input parameters and arguments for the run:")
        logger.info(args)
        dtype_dict = {'plate': str, 'sample': str, 'mpn': float, 'mpn_adj': float, 'upper': float, 'lower': float, 'rarity': float}
        temp_df= pd.DataFrame(columns=dtype_dict.keys()).astype(dtype_dict)
        logger.info("Importing data from {} and {}".format(args.wellmap, args.data))
        if is_file_or_directory(args.data):
            clean_plate = import_plate(platefile=args.data,
                                    plate_name=args.plate_name,
                                    well_name=args.well_name,
                                    col_name=args.col_name,
                                    row_name=args.row_name,
                                    value_name=args.value_name,
                                    zero_padded=args.zero_padded)
            newplate_df = process_plate(args.wellmap, clean_plate, args=args)
            temp_df = pd.concat([temp_df, newplate_df], ignore_index = True)
        else:
            for plate_file in os.listdir(args.data):
                clean_plate = import_plate(platefile=plate_file,
                                        plate_name=args.plate_name,
                                        well_name=args.well_name,
                                        col_name=args.col_name,
                                        row_name=args.row_name,
                                        value_name=args.value_name,
                                        zero_padded=args.zero_padded)
                newplate_df = process_plate(args.wellmap, clean_plate, args=args)
                temp_df = pd.concat([temp_df, newplate_df], ignore_index = True)
        temp_df.to_csv(args.outfile)
        logger.info("Completed MPN estimation. Results are in {}".format(args.outfile))
    except Exception as e:
        logger.exception(e)
        sys.exit(1)



import numpy as np
import pytest
#from scipy.optimize import root_scalar
#from scipy.stats import norm

from micrompn import (
    ptEst_MPN,
    ptEstAdj_MPN,
    logL_MPN,
    logLR_MPN,
    logLRroot_MPN,
    jarvisCI_MPN,
    likeRatioCI_MPN,
    rarity_MPN,
    isMissing,
    checkInputs_mpn,
    mpn,
)

def test_ptEst_MPN():
    # Test case 1: All negative tubes
    positive = np.array([0, 0, 0])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    assert ptEst_MPN(positive, tubes, amount) == 0

    # Test case 2: All positive tubes
    positive = np.array([10, 10, 10])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    assert ptEst_MPN(positive, tubes, amount) == float('inf')

    # Test case 3: General case
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    expected_result = 74.22876396932874
    assert expected_result == pytest.approx(ptEst_MPN(positive, tubes, amount))

def test_ptEstAdj_MPN():
    MPN = 74.22876396932874
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    expected_result = 67.92847966627644

    result = ptEstAdj_MPN(MPN, positive, tubes, amount)

    assert isinstance(result, float)
    assert np.isfinite(result)
    assert np.isclose(result, expected_result)

def test_logL_MPN():
    lambda_ = 74.22876396932874
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    expected_result = -10.254505611153164

    result = logL_MPN(lambda_, positive, tubes, amount)

    assert isinstance(result, float)
    assert np.isclose(result, expected_result)

def test_logLR_MPN():
    lambda_ = 74.22876396932874
    lambda_hat = 73.0
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    expected_result = -0.001695304319877522

    result = logLR_MPN(lambda_, lambda_hat, positive, tubes, amount)

    assert isinstance(result, float)
    assert np.isclose(result, expected_result)

def test_logLRroot_MPN():
    lambda_ = 74.22876396932874
    lambda_hat = 73.0
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    crit_val = 0.95
    expected_result = -0.9516953043198775

    result = logLRroot_MPN(lambda_, lambda_hat, positive, tubes, amount, crit_val)

    assert isinstance(result, float)
    assert np.isclose(result, expected_result)

def test_jarvisCI_MPN():
    MPN = 74.22876396932874
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    conf_level = 0.95

    expected_output = {
        'variance': 902.4173735493894,
        'var_logMPN': 0.1637808007299598,
        'LB': 33.58089913036114,
        'UB': 164.07867398144504
    }

    result = jarvisCI_MPN(MPN, positive, tubes, amount, conf_level)

    assert isinstance(result, dict)
    assert result == expected_output

def test_likeRatioCI_MPN():
    MPN = 74.22876396932874
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])
    conf_level = 0.95

    expected_output = {
        'LB': 32.43393676359959,
        'UB': 152.3661892483937
    }

    result = likeRatioCI_MPN(MPN, positive, tubes, amount, conf_level)

    assert isinstance(result, dict)
    assert np.isclose(result['LB'], expected_output["LB"])
    assert np.isclose(result['UB'], expected_output["UB"])

def test_rarity_MPN():
    MPN = 74.22876396932874
    positive = np.array([10, 5, 1])
    tubes = np.array([10, 10, 10])
    amount = np.array([0.1, 0.01, 0.001])

    expected_output = 0.7705316827363445

    result = rarity_MPN(MPN, positive, tubes, amount)

    assert isinstance(result, float)
    assert np.isclose(result, expected_output)

def test_isMissing():
    x1 = [1, 2, 3, float('nan'), 5]
    x2 = [1, 2, 3, float('inf'), 5]
    x3 = np.array([1, 2, 3, np.inf, 5])

    expected_output1 = True
    expected_output2 = True
    expected_output3 = True

    result1 = isMissing(x1)
    result2 = isMissing(x2)
    result3 = isMissing(x3)

    assert isinstance(result1, bool)
    assert isinstance(result2, bool)
    assert isinstance(result3, bool)
    assert result1 == expected_output1
    assert result2 == expected_output2
    assert result3 == expected_output3

def test_checkInputs_mpn():
    positive = [1, 3, 5, 0, 2]
    tubes = [5, 5, 5, 5, 5]
    amount = [1, 0.1, 0.01, 0.001, 0.0001]
    conf_level = 0.95

    # No exception should be raised for valid inputs
    checkInputs_mpn(positive, tubes, amount, conf_level)

    # Test cases for raising exceptions
    with pytest.raises(ValueError, match="positive, tubes, and amount must be the same length"):
        checkInputs_mpn(positive + [1], tubes, amount, conf_level)
    with pytest.raises(ValueError, match="'positive', 'tubes', and 'amount' cannot have missing values"):
        checkInputs_mpn([1, 3, 5, 0, float('nan')], tubes, amount, conf_level)
    with pytest.raises(TypeError, match="must be real number, not str"):
        checkInputs_mpn(positive, tubes, [1, 0.1, 0.01, 0.001, "abc"], conf_level)
    with pytest.raises(ValueError, match="'tubes' must contain positive whole numbers"):
        checkInputs_mpn(positive, [-1, 3, 5, 0, 2], amount, conf_level)
    with pytest.raises(ValueError, match="'positive' must contain non-negative whole numbers"):
        checkInputs_mpn([-1, 3, 5, 0, 2], tubes, amount, conf_level)
    with pytest.raises(ValueError, match="'amount' must contain positive values"):
        checkInputs_mpn(positive, tubes, [1, 0.1, -0.01, 0.001, 0.0001], conf_level)
    with pytest.raises(ValueError, match="'amount' must be in descending order"):
        checkInputs_mpn(positive, tubes, [1, 0.1, 0.01, 0.1, 0.0001], conf_level)
    with pytest.raises(ValueError, match="'conf_level' must be a float"):
        checkInputs_mpn(positive, tubes, amount, "0.95")
    with pytest.raises(ValueError, match="'conf_level' must be between 0 & 1"):
        checkInputs_mpn(positive, tubes, amount, -0.5)
        checkInputs_mpn(positive, tubes, amount, 1.5)
    with pytest.raises(ValueError, match="more positive tubes than possible"):
        checkInputs_mpn([1, 6, 5, 0, 2], tubes, amount, conf_level)


def test_mpn():
    positive = [10, 5, 1]
    tubes = [10, 10, 10]
    amount = [0.1, 0.01, 0.001]
    conf_level = 0.95
    CI_method = "Jarvis"

    expected_output = {'MPN': 74.22876396932874, 'MPN_adj': 67.92847966627644, 'variance': 902.4173735493894, 'var_log': 0.1637808007299598, 'conf_level': 0.95, 'CI_method': 'Jarvis', 'LB': 33.58089913036114, 'UB': 164.07867398144504, 'RI': 0.7705316827363445}


    result = mpn(positive, tubes, amount, conf_level, CI_method)

    assert isinstance(result, dict)
    assert len(result) == len(expected_output)
    assert np.isclose(result["MPN"], expected_output["MPN"])
    assert np.isclose(result["MPN_adj"], expected_output["MPN_adj"])
    assert np.isclose(result["variance"], expected_output["variance"])
    assert np.isclose(result["var_log"], expected_output["var_log"])
    assert np.isclose(result["conf_level"], expected_output["conf_level"])
    assert result["CI_method"] == expected_output["CI_method"]
    assert np.isclose(result["LB"], expected_output["LB"])
    assert np.isclose(result["UB"], expected_output["UB"])
    assert np.isclose(result["RI"], expected_output["RI"])

    CI_method = "LR"
    expected_output =  {'MPN': 74.22876396932874, 'MPN_adj': 67.92847966627644, 'variance': None, 'var_log': None, 'conf_level': 0.95, 'CI_method': 'LR', 'LB': 32.43393676359959, 'UB': 152.3661892483937, 'RI': 0.7705316827363445}

    result = mpn(positive, tubes, amount, conf_level, CI_method)

    assert isinstance(result, dict)
    assert len(result) == len(expected_output)
    assert np.isclose(result["MPN"], expected_output["MPN"])
    assert np.isclose(result["MPN_adj"], expected_output["MPN_adj"])
    assert result["variance"] is None
    assert result["var_log"] is None
    assert np.isclose(result["conf_level"], expected_output["conf_level"])
    assert result["CI_method"] == expected_output["CI_method"]
    assert np.isclose(result["LB"], expected_output["LB"])
    assert np.isclose(result["UB"], expected_output["UB"])
    assert np.isclose(result["RI"], expected_output["RI"])


"""MicroMPN a Python command line program for automating microbiology most probable 
   number (MPN) estimates in laboratory microplates

    This module of MicroMPN is a transpiled version of a portion of the MPN R package:

    Martine Ferguson and John Ihrie. (2019). MPN: Most Probable Number 
    and Other Microbial Enumeration Techniques. R package version 0.3.0. 
    https://CRAN.R-project.org/package=MPN
"""

import math
from typing import Union, Dict, List

import numpy as np
from scipy.optimize import root_scalar
from scipy.stats import binom, chi2, norm


def ptEst_MPN(positive: np.array, tubes: np.array, amount: np.array) -> float:
    """Estimate MPN from observations of growth at different dilutions by maximizing the likelihood.

     Detailed FDA protocols are `here <https://www.fda.gov/food/laboratory-methods-food/bam-appendix-2-most-probable-number-serial-dilutions>`


    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :return: _description_
    :rtype: float
    """


    def scoreFnc(lambda_, positive, tubes, amount):
        # score function: Find root to maximize likelihood.
        LHS = np.sum((positive * amount) / (1 - np.exp(-lambda_ * amount)))
        RHS = np.sum(tubes * amount)
        return LHS - RHS

    all_negative = np.sum(positive) == 0
    all_positive = np.array_equal(positive, tubes)

    if all_negative:
        return 0
    elif all_positive:
        return float('inf')
    else:
        res = root_scalar(scoreFnc, bracket=[1e-04, 1e+14],
                          args=(positive, tubes, amount),
                          method='brentq', rtol=1e-12, maxiter=10000)
        return res.root


def ptEstAdj_MPN(MPN: float, positive: np.array, tubes: np.array, amount: np.array) -> float:
    """Adjust for bias in MPN estimate.

    Method from:

    Haas CN (1989). "Estimation of microbial densities from dilution count experiments" Applied and Environmental Microbiology 55(8), 1934-1942.
    Salama IA, Koch GG, Tolley DH. (1978) "On the estimation of the most probable number in a serial dilution technique." Communications in Statistics - Theory and Methods, 7(13), 1267-1281.

    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :return: A bias adjusted estimate of MPN
    :rtype: float
    """

    if MPN == 0:
        return 0
    elif np.isinf(MPN):
        return np.nan
    else:
        with np.errstate(divide='ignore', over='ignore', invalid='ignore'):
            # np warning context is used because bias correction method produces overflow values of 
            # cosh and sinh for large MPN that don't appreciably effect the fianl result but are to large to compute
            amount2 = amount ** 2
            amount3 = amount ** 3
            lambda_V = MPN * amount
            one_or_more = 1 - np.exp(-lambda_V)
            zi = tubes * one_or_more
            cosh_term = np.cosh(lambda_V) - 1
            D = np.sum((amount2 * zi) / (2 * cosh_term))
            wi1 = amount2 / (2 * (one_or_more ** 2) * (D ** 3))
            wi2 = (amount3 * zi * np.sinh(lambda_V)) / (cosh_term ** 2)
            wi2[np.isnan(wi2)] = 0  # Inf/Inf is NaN (denominator goes to infinity more quickly)
            wi2 = np.sum(wi2)
            wi3 = amount3 / (one_or_more * cosh_term * (D ** 2))
            wi = wi1 * wi2 - wi3
            return MPN - 0.5 * np.sum(wi * zi * np.exp(-lambda_V))


def logL_MPN(lambda_: float, positive: np.array, tubes: np.array, amount: np.array) -> float:
    """Estimate the Log-likelihood of the MPN estimate

    :param lambda_: MPN
    :type lambda_: float
    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :return: Log-likelihood
    :rtype: float
    """
    # log-likelihood
    lambda_amount = lambda_ * amount
    first_term = positive * np.log(1 - np.exp(-lambda_amount))
    second_term = (tubes - positive) * lambda_amount
    return np.sum(first_term - second_term)


def logLR_MPN(lambda_: float, lambda_hat: float, positive: np.array, tubes: np.array, amount: np.array) -> float:
    """Estimate the Log-likelihood ratio of the MPN estimate

    :param lambda_: MPN
    :type lambda_: float
    :param lambda_hat: MPN estimate
    :type lambda_hat: float
    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :return: Log-likelihood ratio
    :rtype: float
    """
    # log-likelihood ratio
    logL_alt = logL_MPN(lambda_hat, positive, tubes, amount)
    logL_null = logL_MPN(lambda_, positive, tubes, amount)
    return 2 * (logL_alt - logL_null)


def logLRroot_MPN(lambda_: float, lambda_hat: float, positive: np.array, tubes: np.array, amount: np.array, crit_val: float) -> float:
    """Find roots to get Log ratio confidence limits

    :param lambda_: MPN
    :type lambda_: float
    :param lambda_hat: MPN estimate
    :type lambda_hat: float
    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :param crit_val: The critical value for estimating the confidence limits
    :type crit_val: float
    :return: _description_
    :rtype: float
    """
    # Find roots to get LR confidence limits
    log_like_ratio = logLR_MPN(lambda_, lambda_hat, positive, tubes, amount)
    return log_like_ratio - crit_val



def jarvisCI_MPN(MPN: float, positive: np.array, tubes: np.array, amount: np.array, conf_level: float) -> Dict[str, float]:
    """Estimate the confidence interval of the MPN estimate using the method in Jarvis et al. (2010)

    Jarvis B, Wilrich C, Wilrich P-T (2010). "Reconsideration of the derivation of Most Probable Numbers, their standard deviations, confidence bounds and rarity values." Journal of Applied Microbiology, 109, 1660-1667. https://doi.org/10.1111/j.1365-2672.2010.04792.x


    :param MPN: MPN estimate
    :type MPN: float
    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :param conf_level: A scalar value between zero and one for the confidence level. Typically 0.95 (i.e., a 95 percent confidence interval).
    :type conf_level: float
    :return: A dictionary with the following values:
        variance: The estimated variance of the MPN estimate. If all tubes are positive or all negative, variance will be NA.
        var_logMPN: The estimated variance of the natural log of the MPN estimate using the Delta Method (see Jarvis et al.). If all tubes are positive or all negative, var_log will be NA. 
        LB: The lower bound of the confidence interval.
        UB: The upper bound of the confidence interval.
    :rtype: Dict[str, float]
    """
    all_negative = np.sum(positive) == 0
    all_positive = np.array_equal(positive, tubes)
    var_logMPN = np.nan
    variance = np.nan
    sig_level = 1 - conf_level

    if not all_negative and not all_positive:
        exp_term = np.exp(-amount * MPN)
        numer = positive * amount ** 2 * exp_term
        denom = (1 - exp_term) ** 2
        variance = 1 / np.sum(numer / denom)

    if all_negative:
        LB = 0
        UB = np.log(1 / sig_level) / np.sum(amount * tubes)
    elif all_positive:
        def fLB(LB):
            return np.log(1 / sig_level) + np.sum(tubes * np.log(1 - np.exp(-amount * LB)))

        result = root_scalar(fLB, bracket=[1e-02, 1e+03], method='brentq', maxiter=1e+04)
        LB = result.root
        UB = np.inf
    else:
        crit_val = norm.ppf(sig_level / 2, loc=0, scale=1)  # asym. normal
        var_logMPN = variance / (MPN ** 2)  # delta method
        SE_log = np.sqrt(var_logMPN)
        ME_log = crit_val * SE_log
        UB = MPN * np.exp(-ME_log)
        LB = MPN * np.exp(ME_log)

    return {'variance': variance, 'var_logMPN': var_logMPN, 'LB': LB, 'UB': UB}


def likeRatioCI_MPN(MPN: float, positive: np.array, tubes: np.array, amount: np.array, conf_level: float)-> Dict[str, float]:
    """Estimation of the confidence interval by the likelihood ratio method.

    For details see:

    Ridout MS (1994). "A comparison of confidence interval methods for dilution series experiments." Biometrics, 50(1), 289-296.

    :param MPN: MPN estimate
    :type MPN: float
    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :param conf_level: A scalar value between zero and one for the confidence level. Typically 0.95 (i.e., a 95 percent confidence interval).
    :type conf_level: float
    :return: A dictionary containing:
        LB: The lower bound of the confidence interval.
        UB: The upper bound of the confidence interval.
    :rtype: Dict[str, float]
    """
    # Likelihood ratio confidence limits
    # See Ridout (1994) - "A Comparison of CI Methods..."
    all_negative = np.sum(positive) == 0
    all_positive = np.array_equal(positive, tubes)
    crit_val = chi2.ppf(conf_level, df=1, loc=0, scale=1)
    if all_negative or all_positive:
        jarvis_bounds = jarvisCI_MPN(MPN, positive, tubes, amount, conf_level)
        LB = jarvis_bounds['LB']
        UB = jarvis_bounds['UB']
    else:
        result1 = root_scalar(logLRroot_MPN, bracket=[1e-04, MPN], method='brentq', maxiter=10000,
                              args=(MPN, positive, tubes, amount, crit_val))
        result2 = root_scalar(logLRroot_MPN, bracket=[MPN, 5 * MPN], method='brentq', maxiter=10000,
                              args=(MPN, positive, tubes, amount, crit_val))
        LB = result1.root
        UB = result2.root

    return {'LB': LB, 'UB': UB}



def rarity_MPN(MPN: float, positive: np.array, tubes: np.array, amount: np.array) -> float:
    """estimate rarity index from MPN

    If the Rarity Index is less than 1e-04, the experimental results are highly improbable. The researcher may consider running the experiment again and/or changing the dilution levels.

    Jarvis B, Wilrich C, Wilrich P-T (2010). "Reconsideration of the derivation of Most Probable Numbers, their standard deviations, 
    confidence bounds and rarity values." Journal of Applied Microbiology, 109, 1660-1667. https://doi.org/10.1111/j.1365-2672.2010.04792.x


    :param MPN: MPN estimate
    :type MPN: float
    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: np.array
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: np.array
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: np.array
    :return: The rarity index
    :rtype: float
    """

    all_negative = sum(positive) == 0
    all_positive = (positive == tubes).all()

    if all_negative or all_positive:
        return 1
    else:
        probs = 1 - np.exp(-MPN * amount)
        positive_max = np.minimum(np.floor(probs * (tubes + 1)), tubes)
        most_prob_L = binom.pmf(positive_max, tubes, probs)
        actual_L = binom.pmf(positive, tubes, probs)
        rarity = np.prod(actual_L) / np.prod(most_prob_L)
        return rarity

def isMissing(x: Union[list, np.array]) -> bool:
    """Check if any values in a vector are missing or infinite

    :param x: a Vector of values
    :type x: Union[List, np.array]
    :return: true is any value is missing of infinite
    :rtype: bool
    """

    return any([math.isnan(i) or math.isinf(i) for i in x])

def checkInputs_mpn(positive: Union[list, np.array], tubes: Union[list,np.array], amount: Union[list, np.array], conf_level: float) -> None:
    """Validations of input data to the MPN function

    :param positive:  a vector of the count of positive wells at a specific dilution
    :type positive: Union[list, np.array]
    :param tubes:  A vector of the total number of wells at a dilution
    :type tubes: Union[list,np.array]
    :param amount:  A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: Union[list, np.array]
    :param conf_level: The confidence level for estimating upper and lower bounds of MPN
    :type conf_level: float
    :raises ValueError: Any validation failure raises a value error

    """

    l_positive = len(positive)
    l_tubes = len(tubes)
    l_amount = len(amount)
    if l_positive != l_tubes or l_tubes != l_amount:
        raise ValueError("positive, tubes, and amount must be the same length")
    for datavect in [positive, tubes, amount]:
        if isMissing(datavect):
            raise ValueError("'positive', 'tubes', and 'amount' cannot have missing values")
    # for datavect in [positive, tubes, amount]:
    #    if not all(isinstance(i, (int, float)) for i in datavect):
    #        raise TypeError("must be real number, not str")
    if any([t < 1 for t in tubes]) or any([t != round(t) for t in tubes]):
        raise ValueError("'tubes' must contain positive whole numbers")
    if any([p < 0 for p in positive]) or any([p != round(p) for p in positive]):
        raise ValueError("'positive' must contain non-negative whole numbers")
    if any([a <= 0 for a in amount]):
        raise ValueError("'amount' must contain positive values")
    if len(amount) > 1:
        if min([amount[i] - amount[i+1] for i in range(len(amount)-1)]) <= 0:
            raise ValueError("'amount' must be in descending order")
    if not isinstance(conf_level, float):
        raise ValueError("'conf_level' must be a float")
    if conf_level <= 0 or conf_level >= 1:
        raise ValueError("'conf_level' must be between 0 & 1")
    if any([p > t for p,t in zip(positive, tubes)]):
        raise ValueError("more positive tubes than possible")

def mpn(positive: Union[List[int], np.array], tubes: Union[List[int], np.array], amount: Union[List[float], np.array], conf_level: float = 0.95, CI_method: str = "Jarvis") -> dict:
    """mpn calculates the Most Probable Number (MPN) point estimate and confidence interval for microbial concentrations. Also calculates Blodgett's (2002, 2005, 2010) Rarity Index (RI).

    This python code was transpiled from:

    Martine Ferguson and John Ihrie. (2019). MPN: Most Probable Number and Other Microbial Enumeration Techniques. R package version 0.3.0. https://CRAN.R-project.org/package=MPN .
    
    As an example, assume we start with 3g of undiluted inoculum per tube, then use a 10-fold dilution for 2 dilutions. We now have amount = 3 * c(1, .1, .01).
    When all tubes are negative, the point estimate of MPN is zero (same approach as Jarvis et al.). The point estimate for the BAM tables "is listed as less than the lowest MPN for an outcome with at least one positive tube" (App.2).
    When all tubes are positive, the point estimate for MPN is Inf (same approach as Jarvis et al.) since no finite maximum likelihood estimate (MLE) exists. The BAM tables "list the MPN for this outcome as greater than the highest MPN for an outcome with at least one negative tube" (App.2). Here, the difference is probably trivial since the sample should be further diluted if all tubes test positive.
    The bias adjustment for the point estimate uses the method of Salama et al. (1978). Also see Haas (1989).
    Currently, confidence intervals can only be calculated using the Jarvis (2010) or likelihood ratio (LR) approach (Ridout, 1994). The BAM tables use an alternate approach.
    We slightly modified Jarvis' approach when all tubes are positive or all are negative; we use \alpha instead of \alpha / 2α/2 since these are one-sided intervals.
      The Ridout (1994) LR approach uses the same technique (with \alpha) for these two extreme cases.
    If the Rarity Index is less than 1e-04, the experimental results are highly improbable. The researcher may consider running the experiment again and/or changing the dilution levels.

    References:

    Bacteriological Analytical Manual, 8th Edition, Appendix 2, https://www.fda.gov/Food/FoodScienceResearch/LaboratoryMethods/ucm109656.htm

    Blodgett RJ (2002). "Measuring improbability of outcomes from a serial dilution test." Communications in Statistics: Theory and Methods, 31(12), 2209-2223. https://www.tandfonline.com/doi/abs/10.1081/STA-120017222

    Blodgett RJ (2005). "Serial dilution with a confirmation step." Food Microbiology, 22(6), 547-552. https://doi.org/10.1016/j.fm.2004.11.017

    Blodgett RJ (2010). "Does a serial dilution experiment's model agree with its outcome?" Model Assisted Statistics and Applications, 5(3), 209-215. https://doi.org/10.3233/MAS-2010-0157

    Haas CN (1989). "Estimation of microbial densities from dilution count experiments" Applied and Environmental Microbiology 55(8), 1934-1942.

    Haas CN, Rose JB, Gerba CP (2014). "Quantitative microbial risk assessment, Second Ed." John Wiley & Sons, Inc., ISBN 978-1-118-14529-6.

    Jarvis B, Wilrich C, Wilrich P-T (2010). "Reconsideration of the derivation of Most Probable Numbers, their standard deviations, confidence bounds and rarity values." Journal of Applied Microbiology, 109, 1660-1667. https://doi.org/10.1111/j.1365-2672.2010.04792.x

    Ridout MS (1994). "A comparison of confidence interval methods for dilution series experiments." Biometrics, 50(1), 289-296.

    Salama IA, Koch GG, Tolley DH. (1978) "On the estimation of the most probable number in a serial dilution technique." Communications in Statistics - Theory and Methods, 7(13), 1267-1281.



    :param positive: A vector of number of positive tubes at each dilution level.
    :type positive: Union[List, np.array]
    :param tubes: A vector of total number of tubes at each dilution level.
    :type tubes: Union[List,np.array]
    :param amount: A vector of the dilution for each group of wells (smaller is more dilute) 
    :type amount: Union[List, np.array]
    :param conf_level: A scalar value between zero and one for the confidence level. Typically 0.95 (i.e., a 95 percent confidence interval). Defaults to 0.95
    :type conf_level: float, optional
    :param CI_method: The method used for calculating the confidence interval. Choices are "Jarvis" or "LR" (likelihood ratio). Defaults to "Jarvis"
    :type CI_method: str, optional
    :return: A dictionary with these values:
             MPN: The most probable number point estimate for microbial density (concentration).
             MPN_adj: The bias-adjusted point estimate for MPN.
             variance: The estimated variance (see Jarvis et al.) of the MPN estimate if CI_method = "Jarvis". If all tubes are positive or all negative, variance will be NA. If CI_method is not "Jarvis", variance will be NA.
             var_log: The estimated variance of the natural log of the MPN estimate (see Jarvis et al.) using the Delta Method. If all tubes are positive or all negative, var_log will be NA. If CI_method is not "Jarvis", var_log will be NA.
             conf_level: The confidence level used.
             CI_method: The confidence interval method used.
             LB: The lower bound of the confidence interval.
             UB: The upper bound of the confidence interval.
             RI: The rarity index.
    :rtype: Dict[str, float]
    """
    checkInputs_mpn(positive=positive, tubes=tubes, amount=amount, conf_level=conf_level)
    positive_np = np.array(positive)
    tubes_np = np.array(tubes)
    amount_np = np.array(amount)
    MPN = ptEst_MPN(positive=positive_np, tubes=tubes_np, amount=amount_np)
    MPN_adj = ptEstAdj_MPN(MPN=MPN, positive=positive_np, tubes=tubes_np, amount=amount_np)

    variance = None
    var_logMPN = None
    if CI_method == "Jarvis":
        jarvis = jarvisCI_MPN(MPN=MPN, positive=positive_np, tubes=tubes_np, amount=amount_np, conf_level=conf_level)
        variance = jarvis["variance"]
        var_logMPN = jarvis["var_logMPN"]
        LB = jarvis["LB"]
        UB = jarvis["UB"]
    elif CI_method == "LR":
        like_ratio = likeRatioCI_MPN(MPN=MPN, positive=positive_np, tubes=tubes_np, amount=amount_np, conf_level=conf_level)
        LB = like_ratio["LB"]
        UB = like_ratio["UB"]

    rarity = rarity_MPN(MPN, positive_np, tubes_np, amount_np)

    return {"MPN": MPN, "MPN_adj": MPN_adj, "variance": variance, "var_log": var_logMPN,
            "conf_level": conf_level, "CI_method": CI_method, "LB": LB, "UB": UB, "RI": rarity}




import argparse

import pytest
import numpy as np
import pandas as pd


from micrompn import (
    myparser,
    split_dataframe,
    count_above_cutoff,
    trim_counts,
    import_plate,
    process_plate,
    is_file_or_directory,
    main
)

def test_myparser():
    parser = myparser()

    assert isinstance(parser, argparse.ArgumentParser)
    assert parser.description == 'MicroMPN: Software to estimate Most Probable Number (MPN) bacterial abundance from microplates'
    assert any(action.option_strings[0] == '--wellmap' for action in parser._actions)
    assert any(action.option_strings[0] == '--data' for action in parser._actions)
    assert any(action.option_strings[0] == '--cutoff' for action in parser._actions)
    assert any(action.option_strings[0] == '--outfile' for action in parser._actions)
    assert any(action.option_strings[0] == '--plate_name' for action in parser._actions)
    assert any(action.option_strings[0] == '--value_name' for action in parser._actions)
    assert any(action.option_strings[0] == '--well_name' for action in parser._actions)
    assert any(action.option_strings[0] == '--col_name' for action in parser._actions)
    assert any(action.option_strings[0] == '--row_name' for action in parser._actions)
    assert any(action.option_strings[0] == '--zero_padded' for action in parser._actions)
    assert any(action.option_strings[0] == '--trim_positives' for action in parser._actions)
    assert any(action.option_strings[0] == '--version' for action in parser._actions)
    assert any(action.option_strings[0] == '--logfile' for action in parser._actions)
    assert parser._option_string_actions['--version'].dest == 'version'
    assert parser._option_string_actions['--logfile'].dest == 'logfile'

def test_split_dataframe():
    # Create a sample dataframe


    df = pd.DataFrame({
        'plate': ['Plate1', 'Plate1', 'Plate2', 'Plate2'],
        'sample': ['Sample1', 'Sample1', 'Sample2', 'Sample2'],
        'value': [1., 2., 3., 4.]
    })

    # Call the function to split the dataframe
    result_dfs, result_unique_samples = split_dataframe(df)

    # Verify the type and length of the returned values
    assert isinstance(result_dfs, list)
    assert isinstance(result_unique_samples, np.ndarray)
    assert len(result_dfs) == len(result_unique_samples)

    # Verify the content of the split dataframes
    expected_dfs = [
        df[df['plate_sample'] == 'Plate1_Sample1'],
        df[df['plate_sample'] == 'Plate2_Sample2'],
    ]
    for expected_df, result_df in zip(expected_dfs, result_dfs):
        assert result_df.equals(expected_df)

    # Verify the content of the unique samples series
    expected_unique_samples = pd.Series(['Plate1_Sample1', 'Plate2_Sample2'])
    assert pd.Series(result_unique_samples).equals(expected_unique_samples)

def test_count_above_cutoff():
    # Create a sample dataframe
    df = pd.DataFrame({
        'dilution': [1.0, 1.0, 1.0, 0.1, 0.1, 0.1],
        'value': [5.5, 4.5, 3.5, 3.5, 2.5, 1.5]
    })
    cutoff = 3.0

    # Call the function to count positive wells above the cutoff
    result = count_above_cutoff(df, cutoff)

    # Verify the type and length of the returned array
    assert isinstance(result, np.ndarray)
    assert len(result) == len(df['dilution'].unique())

    # Verify the content of the counts array
    expected_counts = np.array([3,1])
    assert np.array_equal(result, expected_counts)

def test_trim_counts():
    # Test case 1: Trim all positive and negative dilutions exist
    positive = np.array([10, 10, 5, 1, 0, 0])
    tubes = np.array([10, 10, 10, 10, 10, 10])
    dilution = np.array([1., 0.1, 0.01, 0.001, 0.0001, 0.00001])
    expected_result = (
        np.array([10, 5, 1, 0]),
        np.array([10, 10, 10, 10]),
        np.array([0.1, 0.01, 0.001, 0.0001])
    )

    result = trim_counts(positive, tubes, dilution)
    print(result)
    assert isinstance(result, tuple)
    assert len(result) == 3
    assert np.array_equal(result[0], expected_result[0])
    assert np.array_equal(result[1], expected_result[1])
    assert np.array_equal(result[2], expected_result[2])

    # Test case 2: No all positive or all negative dilutions
    positive = np.array([1, 0, 1, 0])
    tubes = np.array([10, 10, 10, 10])
    dilution = np.array([1, 2, 3, 4])

    result = trim_counts(positive, tubes, dilution)
    assert isinstance(result, tuple)
    assert len(result) == 3
    assert np.array_equal(result[0], positive)
    assert np.array_equal(result[1], tubes)
    assert np.array_equal(result[2], dilution)

    # Test case 3: Empty arrays
    positive = np.array([])
    tubes = np.array([])
    dilution = np.array([])

    result = trim_counts(positive, tubes, dilution)
    assert isinstance(result, tuple)
    assert len(result) == 3
    assert np.array_equal(result[0], positive)
    assert np.array_equal(result[1], tubes)
    assert np.array_equal(result[2], dilution)
